unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, SyncObjs, StdCtrls, ExtCtrls;

const
  COUNT_READERS = 10;
  READER_DATA_SIZE = 100;

type
  TForm1 = class;
  TSharedData = class;

  ISharedDataAccess = interface
    function  GetElement( AIndex : Integer ) : Real;
    procedure SetElement( AIndex : Integer; AValue : Real );
    function  GetCount() : Integer;

    property Count : Integer read GetCount;
    property Items[ AIndex : Integer ] : Real read GetElement write SetElement; default; 
    
  end;

  TSharedDataAccess = class( TInterfacedObject, ISharedDataAccess )  
  private
    FOwner : TSharedData; 
  
  protected
    constructor Create( AOwner : TSharedData );

  protected
    function  GetElement( AIndex : Integer ) : Real;
    procedure SetElement( AIndex : Integer; AValue : Real );
    function  GetCount() : Integer;

  end;

  TSharedDataRead = class( TSharedDataAccess )
  protected
    constructor Create( AOwner : TSharedData ); 

  public
    destructor  Destroy(); override;
    
  end;

  TSharedDataWrite = class( TSharedDataAccess )
  protected
    constructor Create( AOwner : TSharedData ); 

  public
    destructor  Destroy(); override;
    
  end;

  TSharedData = class
  protected
    FLock : TMultiReadExclusiveWriteSynchronizer;
    FData : array[ 0..9999 ] of Real;

  public
    function GetRead() : ISharedDataAccess;
    function GetWrite() : ISharedDataAccess;

  public
    constructor Create();
    destructor  Destroy(); override;

  end;

  TReaderData = record
    FDataSection : TCriticalSection;
    FData : array[ 0..READER_DATA_SIZE - 1 ] of Real;
  end;

  TReaderThread = class( TThread )
  protected
    FOwner : TForm1;
    FIndex : Integer;
    FData  : TReaderData;

  protected
    procedure Execute(); override;
    
  protected
    procedure ProcessData();

  public
    constructor Create( AOwner : TForm1; AIndex : Integer );
    destructor  Destroy(); override;

  end;

  TWriterThread = class( TThread )
  protected
    FOwner      : TForm1;
    FStartEvent : TEvent;

  protected
    procedure Execute(); override;

  protected
    procedure ProcessData();
    
  public
    procedure Terminate();
    procedure WakeUp();

  public
    constructor Create( AOwner : TForm1 );
    destructor  Destroy(); override;
    
  end;
  
  TForm1 = class(TForm)
    Button1: TButton;
    Timer1: TTimer;
    ListBox1: TListBox;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FData : TSharedData;
    FReaders : array [0..COUNT_READERS - 1] of TReaderThread;
    FWriter  : TWriterThread;
    FDisplayIndex : Integer; 

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses Math;

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
var
  I : Integer;
  
begin
  TThread.NameThreadForDebugging( 'Main', GetCurrentThreadId() );
  FData := TSharedData.Create();
  for I := 0 to COUNT_READERS - 1 do
    FReaders[ I ] := TReaderThread.Create( Self, I );

  FWriter := TWriterThread.Create( Self );
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
  I : Integer;
  
begin
  FWriter.Terminate();
  FWriter.WaitFor();
  FWriter.Free();

  for I := 0 to COUNT_READERS - 1 do
    begin
    FReaders[ I ].Terminate();
    FReaders[ I ].WaitFor();
    FReaders[ I ].Free();
    end;
    
  FData.Free();
end;

constructor TReaderThread.Create( AOwner : TForm1; AIndex : Integer );
begin
  inherited Create( True );
  FOwner := AOwner;
  FIndex := AIndex;
  FData.FDataSection := TCriticalSection.Create();
  Resume();  
end;

destructor TReaderThread.Destroy();
begin
  FData.FDataSection.Free();
  inherited;
end;

procedure TReaderThread.ProcessData();
var
  I : Integer;
  ASharedData : ISharedDataAccess;

begin
  ASharedData := FOwner.FData.GetRead();
  FData.FDataSection.Enter();
  for I := 0 to Length( FData.FData ) - 1 do
    FData.FData[ I ] := ASharedData[ RandomRange( 0, ASharedData.Count ) ];

  FData.FDataSection.Leave();
end;

procedure TReaderThread.Execute();
begin
  NameThreadForDebugging( 'Reader#' + IntToStr( FIndex ), ThreadID );
  while( not Terminated ) do
    begin
    Sleep( 10 );
    ProcessData();
    end;
    
end;

constructor TWriterThread.Create( AOwner : TForm1 );
begin
  inherited Create( True );
  FOwner := AOwner;
  FStartEvent := TEvent.Create( NIL, False, False, '' );
  Resume();  
end;

destructor TWriterThread.Destroy();
begin
  FStartEvent.Free();
  inherited;
end;

procedure TWriterThread.Terminate();
begin
  inherited Terminate();
  FStartEvent.SetEvent();
end;

procedure TWriterThread.WakeUp();
begin
  FStartEvent.SetEvent();
end;

procedure TWriterThread.ProcessData();
var
  I : Integer;
  ASharedData : ISharedDataAccess;
  
begin
  ASharedData := FOwner.FData.GetWrite();

  for I := 0 to ASharedData.Count - 1 do
    ASharedData[ I ] := RandomRange( -1000, 1000 );

end;

procedure TWriterThread.Execute();
begin
  NameThreadForDebugging( 'Writer', ThreadID );
  while( not Terminated ) do
    begin
    FStartEvent.WaitFor( INFINITE );
    ProcessData();
    end;

end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  FWriter.WakeUp();
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  AData : array[ 0..READER_DATA_SIZE - 1 ] of Real;
  I : Integer;

begin
  FReaders[ FDisplayIndex ].FData.FDataSection.Enter();
  Move( FReaders[ FDisplayIndex ].FData.FData[ 0 ], AData[ 0 ], SizeOf( AData )); 
  FReaders[ FDisplayIndex ].FData.FDataSection.Leave();

  Label1.Caption := 'Data from thread #' + IntToStr( FDisplayIndex );
  ListBox1.Items.BeginUpdate();
  ListBox1.Items.Clear(); 
  for I := 0 to READER_DATA_SIZE - 1 do
    ListBox1.Items.Add( FloatToStr( AData[ I ] ));
  
  ListBox1.Items.EndUpdate();

  Inc( FDisplayIndex );
  if( FDisplayIndex >= Length( FReaders ) ) then
    FDisplayIndex := 0;

end;

constructor TSharedDataAccess.Create( AOwner : TSharedData );
begin
  inherited Create();
  FOwner := AOwner;
end;

function TSharedDataAccess.GetElement( AIndex : Integer ) : Real;
begin
  Result := FOwner.FData[ AIndex ];
end;

procedure TSharedDataAccess.SetElement( AIndex : Integer; AValue : Real );
begin
  FOwner.FData[ AIndex ] := AValue;
end;

function TSharedDataAccess.GetCount() : Integer;
begin
  Result := Length( FOwner.FData ); 
end;

constructor TSharedDataRead.Create( AOwner : TSharedData );
begin
  inherited;
  FOwner.FLock.BeginRead();
end;

destructor TSharedDataRead.Destroy();
begin
  FOwner.FLock.EndRead();
  inherited;
end;

constructor TSharedDataWrite.Create( AOwner : TSharedData );
begin
  inherited;
  FOwner.FLock.BeginWrite();
end;

destructor TSharedDataWrite.Destroy();
begin
  FOwner.FLock.EndWrite();
  inherited;
end;

constructor TSharedData.Create();
begin
  inherited;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create();
end;

destructor TSharedData.Destroy();
begin
  FLock.Free();
  inherited;
end;

function TSharedData.GetRead() : ISharedDataAccess;
begin
  Result := TSharedDataRead.Create( Self );
end;

function TSharedData.GetWrite() : ISharedDataAccess;
begin
  Result := TSharedDataWrite.Create( Self );
end;

end.
 